package checkers.util;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Name;

import checkers.types.QualifierHierarchy;

// It's functional, but requires optimization and better documentation
//
/**
 * Represents the type qualifier hierarchy of a type system.
 */
public class DAGQualifierHierarchy extends QualifierHierarchy {

    /**
     * Factory used to create an instance of {@link DAGQualifierHierarchy}.
     *
     * Clients build the hierarchy using {@link #addQualifier(AnnotationMirror)}
     * and {@link #addSubtype(AnnotationMirror, AnnotationMirror)}, then get
     * the instance with calling {@link #build()}
     */
    public static class Factory {
        /** map: qualifier --> supertypes of the qualifier */
        private Map<AnnotationMirror, Set<AnnotationMirror>> subtypes;

        /**
         * Creates a factory for a hierarchy that initalized to be a copy of
         * the passed hierarchy.
         */
        public Factory(DAGQualifierHierarchy hierarchy) {
            subtypes = copy(hierarchy.subtypes);
        }

        public Factory() {
            subtypes = AnnotationUtils.createAnnotationMap();
        }

        /**
         * Adds the passed qualifier to the hierarchy.  Clients need to specify
         * its super qualifiers in subsequent calls to
         * {@link #addSubtype(AnnotationMirror, AnnotationMirror)}.
         */
        public void addQualifier(AnnotationMirror qual) {
            if (subtypes.containsKey(qual))
                return;
            subtypes.put(qual, AnnotationUtils.createAnnotationSet());
        }

        /**
         * Adds a subtype relationship between the two type qualifiers.
         *
         * @param sub   the sub type qualifier
         * @param sup   the super type qualifier
         */
        public void addSubtype(AnnotationMirror sub, AnnotationMirror sup) {
            addQualifier(sub);
            addQualifier(sup);
            subtypes.get(sub).add(sup);
        }

        /**
         * Returns an instance of {@link DAGQualifierHierarchy} that represent
         * the hierarchy built so far
         */
        public DAGQualifierHierarchy build() {
            return new DAGQualifierHierarchy(this);
        }
    }

    /** immutable map: qualifier --> supertypes of the qualifier**/
    private final Map<AnnotationMirror, Set<AnnotationMirror>> subtypes;
    /** the root of all the qualifiers **/
    private final AnnotationMirror root;

    private DAGQualifierHierarchy(Factory f) {
        this.subtypes = copyAndConst(f.subtypes);
        this.root = findRoot(this.subtypes);
    }

    protected DAGQualifierHierarchy(DAGQualifierHierarchy h) {
        this.subtypes = h.subtypes;
        this.root = h.root;
    }

    /**
     * Returns the root qualifier for this hierarchy.
     *
     * The root qualifier is inferred from the hierarchy, as being the only
     * one without any super qualifiers
     */
    @Override
    public AnnotationMirror getRootAnnotation() {
        return root;
    }

    @Override
    public Set<Name> getTypeQualifiers() {
        Set<Name> names = new HashSet<Name>();
        for (AnnotationMirror anno: subtypes.keySet())
            names.add(AnnotationUtils.annotationName(anno));
        return Collections.unmodifiableSet(names);
    }

    @Override
    public AnnotationMirror leastUpperBound(AnnotationMirror a1, AnnotationMirror a2) {
        if (AnnotationUtils.areSame(a1, a2))
            return a1;
        if (AnnotationUtils.areSame(root, a1) || AnnotationUtils.areSame(root, a2))
            return root;
        if (isSubtype(a1, a2))
            return a2;
        else if (isSubtype(a2, a1))
            return a1;
        return root;
    }

    //
    // These subtype query methods have an "interesting" specification for
    // dealing with annotation with values.  Let's take IGJ @I as example
    // to illustrate the specification
    //
    // Two annotations with values are subtype of each other only if they
    // have the same values.  i.e. I(m) is a subtype of I(n) iff m = n
    //
    // When use specify an annotation, a1, to be a subtype of annotation with
    // values, a2, then a1 is a subtype of all instances of a2 regardless of
    // a2 values.  i.e. IGJBottom is a subtype of all instances of @I
    //
    @Override
    public boolean isSubtype(AnnotationMirror anno1, AnnotationMirror anno2) {
        if (AnnotationUtils.areSame(root, anno2))
            return true;
        if (AnnotationUtils.areSameIgnoringValues(anno1, anno2))
            return AnnotationUtils.areSame(anno1, anno2);
        return isSubtypeImp(anno1, anno2);
    }

    private boolean isSubtypeImp(AnnotationMirror anno1, AnnotationMirror anno2) {
        if (AnnotationUtils.areSameIgnoringValues(anno1, anno2))
            return true;
        Set<AnnotationMirror> a1Supers = subtypes.get(anno1);
        assert a1Supers != null : anno1 + " qualifier is not in the hierarchy";
        if (a1Supers.isEmpty())
            return false;
        for (AnnotationMirror a1Super : a1Supers) {
            if (isSubtypeImp(a1Super, anno2))
                return true;
        }
        return false;
    }

    /**
     * Creates a deep copy of the subtypes map
     */
    private static Map<AnnotationMirror, Set<AnnotationMirror>>
    copy(Map<AnnotationMirror, Set<AnnotationMirror>> subtypes) {
        Map<AnnotationMirror, Set<AnnotationMirror>> copy = AnnotationUtils.createAnnotationMap();

        for (AnnotationMirror key : subtypes.keySet()) {
            Set<AnnotationMirror> korig = subtypes.get(key);
            Set<AnnotationMirror> kcopy = AnnotationUtils.createAnnotationSet();
            kcopy.addAll(korig);
            copy.put(key, kcopy);
        }
        return copy;
    }

    /**
     * Creates an immutable deep copy of subtypes map
     */
    private static Map<AnnotationMirror, Set<AnnotationMirror>>
    copyAndConst(Map<AnnotationMirror, Set<AnnotationMirror>> subtypes) {
        Map<AnnotationMirror, Set<AnnotationMirror>> copy = AnnotationUtils.createAnnotationMap();

        for (AnnotationMirror key : subtypes.keySet()) {
            Set<AnnotationMirror> korig = subtypes.get(key);
            Set<AnnotationMirror> kcopy = AnnotationUtils.createAnnotationSet();
            kcopy.addAll(korig);
            copy.put(key, Collections.unmodifiableSet(kcopy));
        }
        return Collections.unmodifiableMap(copy);
    }


    /**
     * Infer the root for the subtype hierarhy.  Simply finds the one (and only
     * one) qualifier that is not a subtype of any other qualifier
     */
    private static AnnotationMirror
    findRoot(Map<AnnotationMirror, Set<AnnotationMirror>> subtypes) {
        List<AnnotationMirror> possibleRoots = new LinkedList<AnnotationMirror>();
        for (AnnotationMirror anno : subtypes.keySet()) {
            if (subtypes.get(anno).isEmpty())
                possibleRoots.add(anno);
        }
        assert possibleRoots.size() == 1 : possibleRoots;
        return possibleRoots.get(0);
    }
}
